/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Piotr Gregor <piotr@freeswitch.com> <piotr@dataandsignal.com>
 *
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Piotr Gregor <piotr@freeswitch.org> <piotr@signalwire.com>
 * Dragos Oancea <dragos@signalwire.com>
 *
 * mod_red.c -- The redundancy encoder/decoder (RFC 2198, RFC 5109)
 *
 */

#include <switch.h>
#include <switch_ulp.h>

uint32_t ulp_binomial_table[(1 + ULP_MAX_FEC_IN_SCHEME) * (1 + ULP_MAX_FEC_IN_SCHEME + 1) / 2] = {
	1,				/* n=0, k=0 */
	1, 1,			/* n=1, k={0,1} */
	1, 2, 1,		/* n=2, k={0,1,2} */
	1, 3, 3, 1,
	1, 4, 6, 4, 1,
	1, 5, 10, 10, 5, 1,
	1, 6, 15, 20, 15, 6, 1,
	1, 7, 21, 35, 35, 21, 7, 1, 
	1, 8, 28, 56, 70, 56, 28, 8, 1,
	1, 9, 36, 84, 126, 126, 84, 36,9, 1,
	1, 10, 45, 120, 210, 252, 210, 120, 45, 10, 1,
	1, 11, 55, 165, 330, 462, 462, 330, 165, 55, 11, 1,
	1, 12, 66, 220, 495, 792, 924, 792, 495, 220, 66, 12, 1 
};

/* Protection masks. */

static uint16_t ulp_mask16_1_1[1] = {
	0x8000
};

static uint16_t ulp_mask16_2_1[1] = {
	0xC000
};

static uint16_t ulp_mask16_2_2[2] = {
	0x8000, 0xC000
};

static uint16_t ulp_mask16_3_1[1] = {
  0xe000
};

static uint16_t ulp_mask16_3_2[2] = {
  0xc000,
  0xa000
};

static uint16_t ulp_mask16_3_3[3] = {
  0x8000,
  0xc000,
  0x6000
};

static uint16_t ulp_mask16_4_1[] = {
  0xf000
};

static uint16_t ulp_mask16_4_2[] = {
  0xa000,
  0xd000
};

static uint16_t ulp_mask16_4_3[] = {
  0xc000,
  0x6000,
  0x9000
};

static uint16_t ulp_mask16_4_4[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000
};

static uint16_t ulp_mask16_5_1[] = {
  0xf800
};

static uint16_t ulp_mask16_5_2[] = {
  0xd000,
  0xa800
};

static uint16_t ulp_mask16_5_3[] = {
  0x7000,
  0x9000,
  0xc800
};

static uint16_t ulp_mask16_5_4[] = {
  0xc000,
  0x6000,
  0x3000,
  0x8800
};

static uint16_t ulp_mask16_5_5[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800
};

static uint16_t ulp_mask16_6_1[] = {
  0xfc00
};

static uint16_t ulp_mask16_6_2[] = {
  0xa800,
  0xd400
};

static uint16_t ulp_mask16_6_3[] = {
  0x9400,
  0xc800,
  0x6400
};

static uint16_t ulp_mask16_6_4[] = {
  0x6000,
  0x3800,
  0x8800,
  0xc400
};

static uint16_t ulp_mask16_6_5[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x8400
};

static uint16_t ulp_mask16_6_6[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00
};

static uint16_t ulp_mask16_7_1[] = {
  0xfe00
};

static uint16_t ulp_mask16_7_2[] = {
  0xd400,
  0xaa00
};

static uint16_t ulp_mask16_7_3[] = {
  0xc800,
  0x7400,
  0x9200
};

static uint16_t ulp_mask16_7_4[] = {
  0x3800,
  0x8a00,
  0xc400,
  0x6200
};

static uint16_t ulp_mask16_7_5[] = {
  0x6000,
  0x3000,
  0x1c00,
  0x8400,
  0xc200
};

static uint16_t ulp_mask16_7_6[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x8200
};

static uint16_t ulp_mask16_7_7[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600
};

static uint16_t ulp_mask16_8_1[] = {
  0xff00
};

static uint16_t ulp_mask16_8_2[] = {
  0xaa00,
  0xd500
};

static uint16_t ulp_mask16_8_3[] = {
  0x7400,
  0x9200,
  0xc900
};

static uint16_t ulp_mask16_8_4[] = {
  0x8a00,
  0xc500,
  0x6200,
  0x3100
};

static uint16_t ulp_mask16_8_5[] = {
  0x3000,
  0x1c00,
  0x8500,
  0xc200,
  0x6100
};

static uint16_t ulp_mask16_8_6[] = {
  0x6000,
  0x3000,
  0x1800,
  0x0e00,
  0x8200,
  0xc100
};

static uint16_t ulp_mask16_8_7[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x8100
};

static uint16_t ulp_mask16_8_8[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300
};

static uint16_t ulp_mask16_9_1[] = {
  0xff80
};

static uint16_t ulp_mask16_9_2[] = {
  0xd500,
  0xaa80
};

static uint16_t ulp_mask16_9_3[] = {
  0x9200,
  0xc900,
  0x7480
};

static uint16_t ulp_mask16_9_4[] = {
  0xc500,
  0x6200,
  0x3900,
  0x8a80
};

static uint16_t ulp_mask16_9_5[] = {
  0x1c00,
  0x8500,
  0xc280,
  0x6100,
  0x3080
};

static uint16_t ulp_mask16_9_6[] = {
  0x3000,
  0x1800,
  0x0e00,
  0x8280,
  0xc100,
  0x6080
};

static uint16_t ulp_mask16_9_7[] = {
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0700,
  0x8100,
  0xc080
};

static uint16_t ulp_mask16_9_8[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x8080
};

static uint16_t ulp_mask16_9_9[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180
};

static uint16_t ulp_mask16_10_1[] = {
  0xffc0
};

static uint16_t ulp_mask16_10_2[] = {
  0xaa80,
  0xd540
};

static uint16_t ulp_mask16_10_3[] = {
  0xc900,
  0x7480,
  0x9240
};

static uint16_t ulp_mask16_10_4[] = {
  0x6200,
  0x3900,
  0x8a80,
  0xc540
};

static uint16_t ulp_mask16_10_5[] = {
  0x8500,
  0xc280,
  0x6140,
  0x3080,
  0x1840
};

static uint16_t ulp_mask16_10_6[] = {
  0x1800,
  0x0e00,
  0x8280,
  0xc140,
  0x6080,
  0x3040
};

static uint16_t ulp_mask16_10_7[] = {
  0x3000,
  0x1800,
  0x0c00,
  0x0700,
  0x8140,
  0xc080,
  0x6040
};

static uint16_t ulp_mask16_10_8[] = {
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x8080,
  0xc040
};

static uint16_t ulp_mask16_10_9[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x8040
};

static uint16_t ulp_mask16_10_10[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x00c0
};

static uint16_t ulp_mask16_11_1[] = {
  0xffe0
};

static uint16_t ulp_mask16_11_2[] = {
  0xd540,
  0xaaa0
};

static uint16_t ulp_mask16_11_3[] = {
  0x7480,
  0x9240,
  0xc920
};

static uint16_t ulp_mask16_11_4[] = {
  0x3900,
  0x8a80,
  0xc540,
  0x6220
};

static uint16_t ulp_mask16_11_5[] = {
  0xc2c0,
  0x6100,
  0x30a0,
  0x1c40,
  0x8520
};

static uint16_t ulp_mask16_11_6[] = {
  0x0e00,
  0x8280,
  0xc140,
  0x60a0,
  0x3040,
  0x1820
};

static uint16_t ulp_mask16_11_7[] = {
  0x1800,
  0x0c00,
  0x0700,
  0x8140,
  0xc0a0,
  0x6040,
  0x3020
};

static uint16_t ulp_mask16_11_8[] = {
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0340,
  0x80a0,
  0xc040,
  0x6020
};

static uint16_t ulp_mask16_11_9[] = {
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x8040,
  0xc020
};

static uint16_t ulp_mask16_11_10[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x00c0,
  0x8020
};

static uint16_t ulp_mask16_11_11[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x00c0,
  0x0060
};

static uint16_t ulp_mask16_12_1[] = {
  0xfff0
};

static uint16_t ulp_mask16_12_2[] = {
  0xaaa0,
  0xd550
};

static uint16_t ulp_mask16_12_3[] = {
  0x9240,
  0xc920,
  0x7490
};

static uint16_t ulp_mask16_12_4[] = {
  0x8a80,
  0xc540,
  0x6220,
  0x3910
};

static uint16_t ulp_mask16_12_5[] = {
  0x6100,
  0x30a0,
  0x1c50,
  0x8520,
  0xc290
};

static uint16_t ulp_mask16_12_6[] = {
  0x8290,
  0xc140,
  0x60a0,
  0x3050,
  0x1820,
  0x0c10
};

static uint16_t ulp_mask16_12_7[] = {
  0x0c00,
  0x0700,
  0x8140,
  0xc0a0,
  0x6050,
  0x3020,
  0x1810
};

static uint16_t ulp_mask16_12_8[] = {
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x80a0,
  0xc050,
  0x6020,
  0x3010
};

static uint16_t ulp_mask16_12_9[] = {
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x8050,
  0xc020,
  0x6010
};

static uint16_t ulp_mask16_12_10[] = {
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x00c0,
  0x8020,
  0xc010
};

static uint16_t ulp_mask16_12_11[] = {
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x00c0,
  0x0060,
  0x8010
};

static uint16_t ulp_mask16_12_12[] = {
  0x8000,
  0xc000,
  0x6000,
  0x3000,
  0x1800,
  0x0c00,
  0x0600,
  0x0300,
  0x0180,
  0x00c0,
  0x0060,
  0x0030
};

/**
 * Protection masks for all FEC schemes up to ULP_MAX_FEC_IN_SCHEME
 * number of FECs in scheme.
 */
uint16_t* ulp_masks16[] = {
	[0] = ulp_mask16_1_1,
	[1] = ulp_mask16_2_1,
	[2] = ulp_mask16_2_2,
	[3] = ulp_mask16_3_1,
	[4] = ulp_mask16_3_2,
	[5] = ulp_mask16_3_3,
	[6] = ulp_mask16_4_1,
	[7] = ulp_mask16_4_2,
	[8] = ulp_mask16_4_3,
	[9] = ulp_mask16_4_4,
	[10] = ulp_mask16_5_1,
	[11] = ulp_mask16_5_2,
	[12] = ulp_mask16_5_3,
	[13] = ulp_mask16_5_4,
	[14] = ulp_mask16_5_5,
	[15] = ulp_mask16_6_1,
	[16] = ulp_mask16_6_2,
	[17] = ulp_mask16_6_3,
	[18] = ulp_mask16_6_4,
	[19] = ulp_mask16_6_5,
	[20] = ulp_mask16_6_6,
	[21] = ulp_mask16_7_1,
	[22] = ulp_mask16_7_2,
	[23] = ulp_mask16_7_3,
	[24] = ulp_mask16_7_4,
	[25] = ulp_mask16_7_5,
	[26] = ulp_mask16_7_6,
	[27] = ulp_mask16_7_7,
	[28] = ulp_mask16_8_1,
	[29] = ulp_mask16_8_2,
	[30] = ulp_mask16_8_3,
	[31] = ulp_mask16_8_4,
	[32] = ulp_mask16_8_5,
	[33] = ulp_mask16_8_6,
	[34] = ulp_mask16_8_7,
	[35] = ulp_mask16_8_8,
	[36] = ulp_mask16_9_1,
	[37] = ulp_mask16_9_2,
	[38] = ulp_mask16_9_3,
	[39] = ulp_mask16_9_4,
	[40] = ulp_mask16_9_5,
	[41] = ulp_mask16_9_6,
	[42] = ulp_mask16_9_7,
	[43] = ulp_mask16_9_8,
	[44] = ulp_mask16_9_9,
	[45] = ulp_mask16_10_1,
	[46] = ulp_mask16_10_2,
	[47] = ulp_mask16_10_3,
	[48] = ulp_mask16_10_4,
	[49] = ulp_mask16_10_5,
	[50] = ulp_mask16_10_6,
	[51] = ulp_mask16_10_7,
	[52] = ulp_mask16_10_8,
	[53] = ulp_mask16_10_9,
	[54] = ulp_mask16_10_10,
	[55] = ulp_mask16_11_1,
	[56] = ulp_mask16_11_2,
	[57] = ulp_mask16_11_3,
	[58] = ulp_mask16_11_4,
	[59] = ulp_mask16_11_5,
	[60] = ulp_mask16_11_6,
	[61] = ulp_mask16_11_7,
	[62] = ulp_mask16_11_8,
	[63] = ulp_mask16_11_9,
	[64] = ulp_mask16_11_10,
	[65] = ulp_mask16_11_11,
	[66] = ulp_mask16_12_1,
	[67] = ulp_mask16_12_2,
	[68] = ulp_mask16_12_3,
	[69] = ulp_mask16_12_4,
	[70] = ulp_mask16_12_5,
	[71] = ulp_mask16_12_6,
	[72] = ulp_mask16_12_7,
	[73] = ulp_mask16_12_8,
	[74] = ulp_mask16_12_9,
	[75] = ulp_mask16_12_10,
	[76] = ulp_mask16_12_11,
	[77] = ulp_mask16_12_12
};

uint16_t* ulp_burst_high_loss_masks16[] = {
	[0] = ulp_mask16_1_1,
	[1] = ulp_mask16_2_2,
	[2] = ulp_mask16_3_3,
	[3] = ulp_mask16_4_4,
	[4] = ulp_mask16_5_5,
	[5] = ulp_mask16_6_6,
	[6] = ulp_mask16_7_7,
	[7] = ulp_mask16_8_8,
	[8] = ulp_mask16_9_9,
	[9] = ulp_mask16_10_10,
	[10] = ulp_mask16_11_11,
	[11] = ulp_mask16_12_12,
};

uint16_t* ulp_burst_medium_loss_masks16[] = {
	[0] = ulp_mask16_1_1,
	[1] = ulp_mask16_2_1,
	[2] = ulp_mask16_3_2,
	[3] = ulp_mask16_4_2,
	[4] = ulp_mask16_5_3,
	[5] = ulp_mask16_6_3,
	[6] = ulp_mask16_7_4,
	[7] = ulp_mask16_8_4,
	[8] = ulp_mask16_9_5,
	[9] = ulp_mask16_10_5,
	[10] = ulp_mask16_11_6,
	[11] = ulp_mask16_12_6,
};

uint16_t* ulp_burst_small_loss_masks16[] = {
	[0] = ulp_mask16_1_1,
	[1] = ulp_mask16_2_2,
	[2] = ulp_mask16_3_2,
	[3] = ulp_mask16_4_2,
	[4] = ulp_mask16_5_2,
	[5] = ulp_mask16_6_2,
	[6] = ulp_mask16_7_2,
	[7] = ulp_mask16_8_2,
	[8] = ulp_mask16_9_2,
	[9] = ulp_mask16_10_2,
	[10] = ulp_mask16_11_2,
	[11] = ulp_mask16_12_2,
};

mask_set_t ulp_burst_small_loss_masks16_set[] = {
	 {ulp_mask16_1_1, 1, 1},
	 {ulp_mask16_2_1, 2, 1},
	 {ulp_mask16_3_2, 3, 2},
	 {ulp_mask16_4_2, 4, 2},
	 {ulp_mask16_5_2, 5, 2},
	 {ulp_mask16_6_2, 6, 2},
	 {ulp_mask16_7_2, 7, 2},
	 {ulp_mask16_8_2, 8, 2},
	 {ulp_mask16_9_2, 9, 2},
	 {ulp_mask16_10_2, 10, 2},
	 {ulp_mask16_11_2, 11, 2},
	 {ulp_mask16_12_2, 12, 2}
}; 

mask_set_t ulp_burst_medium_loss_masks16_set[] = {
	 {ulp_mask16_1_1, 1, 1},
	 {ulp_mask16_2_1, 2, 1},
	 {ulp_mask16_3_2, 3, 2},
	 {ulp_mask16_4_2, 4, 2},
	 {ulp_mask16_5_3, 5, 3},
	 {ulp_mask16_6_3, 6, 3},
	 {ulp_mask16_7_4, 7, 4},
	 {ulp_mask16_8_4, 8, 4},
	 {ulp_mask16_9_5, 9, 5},
	 {ulp_mask16_10_5, 10, 5},
	 {ulp_mask16_11_6, 11, 6},
	 {ulp_mask16_12_6, 12, 6}
}; 

mask_set_t ulp_burst_high_loss_masks16_set[] = {
	 {ulp_mask16_1_1, 1, 1},
	 {ulp_mask16_2_2, 2, 2},
	 {ulp_mask16_3_3, 3, 3},
	 {ulp_mask16_4_4, 4, 4},
	 {ulp_mask16_5_5, 5, 5},
	 {ulp_mask16_6_6, 6, 6},
	 {ulp_mask16_7_7, 7, 7},
	 {ulp_mask16_8_8, 8, 8},
	 {ulp_mask16_9_9, 9, 9},
	 {ulp_mask16_10_10, 10, 10},
	 {ulp_mask16_11_11, 11, 11},
	 {ulp_mask16_12_12, 12, 12}
}; 

uint16_t** ulp_burst_loss_masks_sets_list[] = {
	/* 0 - smallest loss, 9 - highest loss */
	[0] = ulp_burst_small_loss_masks16,
	[1] = ulp_burst_small_loss_masks16,
	[2] = ulp_burst_small_loss_masks16,
	[3] = ulp_burst_medium_loss_masks16,
	[4] = ulp_burst_medium_loss_masks16,
	[5] = ulp_burst_medium_loss_masks16,
	[6] = ulp_burst_medium_loss_masks16,
	[7] = ulp_burst_high_loss_masks16,
	[8] = ulp_burst_high_loss_masks16,
	[9] = ulp_burst_high_loss_masks16
};

mask_set_t* ulp_burst_loss_masks_sets_list_set[] = {
	[0] = (mask_set_t*)&ulp_burst_small_loss_masks16_set,
	[1] = (mask_set_t*)&ulp_burst_small_loss_masks16_set,
	[2] = (mask_set_t*)&ulp_burst_small_loss_masks16_set,
	[3] = (mask_set_t*)&ulp_burst_medium_loss_masks16_set,
	[4] = (mask_set_t*)&ulp_burst_medium_loss_masks16_set,
	[5] = (mask_set_t*)&ulp_burst_medium_loss_masks16_set,
	[6] = (mask_set_t*)&ulp_burst_medium_loss_masks16_set,
	[7] = (mask_set_t*)&ulp_burst_high_loss_masks16_set,
	[8] = (mask_set_t*)&ulp_burst_high_loss_masks16_set,
	[9] = (mask_set_t*)&ulp_burst_high_loss_masks16_set
};

/* Random loss protection masks, for 100 % overhead  */
static uint16_t ulp_rand_high_loss_mask16_1_1[1] = {
	0x8000
};
 
static uint16_t ulp_rand_high_loss_mask16_2_2[2] = {
	0xc000, 
	0x8000
};

static uint16_t ulp_rand_high_loss_mask16_3_3[3] = {
	0xc000,
	0xa000,
	0x6000
};

static uint16_t ulp_rand_high_loss_mask16_4_4[4] = {
	0xc000,
	0xa000,
	0x3000,
	0x5000
};

static uint16_t ulp_rand_high_loss_mask16_5_5[5] = {
	0xc000,
	0x3000, 
	0x1800, 
	0xa000, 
	0x4800
};

static uint16_t ulp_rand_high_loss_mask16_6_6[6] = {
	0x9800, 
	0x6400, 
	0x5000, 
	0x1400, 
	0xa800, 
	0xe000
};

static uint16_t ulp_rand_high_loss_mask16_7_7[7] = {
	0x4a00, 
	0x9400, 
	0x1a00, 
	0xc400, 
	0x2800, 
	0xc200, 
	0x3400
};

static uint16_t ulp_rand_high_loss_mask16_8_8[8] = {
	0x2500, 
	0x8a00, 
	0x9100, 
	0x6800, 
	0x3200, 
	0x4300, 
	0xc400, 
	0x1c00
};

static uint16_t ulp_rand_high_loss_mask16_9_9[9] = {
	0x4c00,
	0x6200, 
	0x9100, 
	0x4280, 
	0xa400, 
	0x1300, 
	0x3080, 
	0x8880, 
	0x0900
};

static uint16_t ulp_rand_high_loss_mask16_10_10[10] = {
	0x4c00, 
	0x2300, 
	0x88c0, 
	0x2140, 
	0x5280, 
	0x9400, 
	0x2600, 
	0x4840, 
	0x9180,
	0xc000 /*!*/
};

static uint16_t ulp_rand_high_loss_mask16_11_11[11] = {
	0x4640,
	0x3320,
	0x9900, 
	0x0580, 
	0x80a0, 
	0x8440, 
	0x4060, 
	0x0a80, 
	0x6800, 
	0x1020, 
	0x3040
};

static uint16_t ulp_rand_high_loss_mask16_12_12[12] = {
	0x1060, 
	0x0230, 
	0x4050, 
	0x2180, 
	0x8110, 
	0x1480, 
	0x9800, 
	0x0890, 
	0x6200, 
	0x2420, 
	0x8a00, 
	0x8440
};

uint16_t* ulp_rand_high_loss_masks16[] = {
	[0] = ulp_rand_high_loss_mask16_1_1,
	[1] = ulp_rand_high_loss_mask16_2_2,
	[2] = ulp_rand_high_loss_mask16_3_3,
	[3] = ulp_rand_high_loss_mask16_4_4,
	[4] = ulp_rand_high_loss_mask16_5_5,
	[5] = ulp_rand_high_loss_mask16_6_6,
	[6] = ulp_rand_high_loss_mask16_7_7,
	[7] = ulp_rand_high_loss_mask16_8_8,
	[8] = ulp_rand_high_loss_mask16_9_9,
	[9] = ulp_rand_high_loss_mask16_10_10,
	[10] = ulp_rand_high_loss_mask16_11_11,
	[11] = ulp_rand_high_loss_mask16_12_12,
};

/* odd number of packets in the frame : we use the masks sets with more fec packets (+1)*/
static uint16_t ulp_rand_medium_loss_mask16_1_1[1] = {
	0x8000 
};

static uint16_t ulp_rand_medium_loss_mask16_2_1[1] = {
	0xc000 
};

static uint16_t ulp_rand_medium_loss_mask16_3_2[2] = {
	0xc000,
	0xa000
};

static uint16_t ulp_rand_medium_loss_mask16_4_2[2] = {
	0xc000,
	0xb000
};

static uint16_t ulp_rand_medium_loss_mask16_5_3[3] = {
	0xb000,
	0xc800,
	0x5000
};

static uint16_t ulp_rand_medium_loss_mask16_6_3[3] = {
	0xd000, 
	0x6800, 
	0xa400
};

static uint16_t ulp_rand_medium_loss_mask16_7_4[4] = {
	0xd000,
	0xaa00, 
	0x6400,
	0x1c00
};

static uint16_t ulp_rand_medium_loss_mask16_8_4[4] = {
	0x4500,
	0xb400, 
	0x6a00, 
	0x8900
};

static uint16_t ulp_rand_medium_loss_mask16_9_5[5] = {
	0x8c00,
	0x2500,
	0x9280,
	0x4180,
	0x5800
};

static uint16_t ulp_rand_medium_loss_mask16_10_5[5] = {
	0xca00,
	0x3280,
	0xa140,
	0x5500,
	0x08c0
};

static uint16_t ulp_rand_medium_loss_mask16_11_6[6] = {
	0x6440, 
	0x5140, 
	0x0ca0, 
	0xa120, 
	0x12a0, 
	0x8a40
};

static uint16_t ulp_rand_medium_loss_mask16_12_6[6] = {
	0x5140, 
	0xc510, 
	0x2180, 
	0x1230, 
	0x08e0, 
	0x2e00
};

uint16_t* ulp_rand_medium_loss_masks16[] = {
	[0] = ulp_rand_medium_loss_mask16_1_1,
	[1] = ulp_rand_medium_loss_mask16_2_1,
	[2] = ulp_rand_medium_loss_mask16_3_2,
	[3] = ulp_rand_medium_loss_mask16_4_2,
	[4] = ulp_rand_medium_loss_mask16_5_3,
	[5] = ulp_rand_medium_loss_mask16_6_3,
	[6] = ulp_rand_medium_loss_mask16_7_4,
	[7] = ulp_rand_medium_loss_mask16_8_4,
	[8] = ulp_rand_medium_loss_mask16_9_5,
	[9] = ulp_rand_medium_loss_mask16_10_5,
	[10] = ulp_rand_medium_loss_mask16_11_6,
	[11] = ulp_rand_medium_loss_mask16_12_6,
};

static uint16_t ulp_rand_small_loss_mask16_1_1[1] = {
	0x8000
};

static uint16_t ulp_rand_small_loss_mask16_2_1[1] = {
	0xc000
};

static uint16_t ulp_rand_small_loss_mask16_3_2[2] = {
	0xc000,
	0xa000
};

static uint16_t ulp_rand_small_loss_mask16_4_2[2] = {
	0xc000,
	0xb000
};

static uint16_t ulp_rand_small_loss_mask16_5_2[2] = {
	0xa800,
	0xd000
};

static uint16_t ulp_rand_small_loss_mask16_6_2[2] = {
	0xa800,
	0xd400
};

static uint16_t ulp_rand_small_loss_mask16_7_2[2] = {
	0xd400,
	0xaa00
};

static uint16_t ulp_rand_small_loss_mask16_8_2[2] = {
	0xaa00,
	0xd500
};

static uint16_t ulp_rand_small_loss_mask16_9_2[2] = {
	0xaa80,
	0xd500
};

static uint16_t ulp_rand_small_loss_mask16_10_2[2] = {
	0xaa80,
	0xd540
};

static uint16_t ulp_rand_small_loss_mask16_11_2[2] = {
	0xecc0,
	0x9ba0
};

static uint16_t ulp_rand_small_loss_mask16_12_2[2] = {
	0xecc0,
	0x93b0
};

uint16_t* ulp_rand_small_loss_masks16[] = {
	[0] = ulp_rand_small_loss_mask16_1_1,
	[1] = ulp_rand_small_loss_mask16_2_1,
	[2] = ulp_rand_small_loss_mask16_3_2,
	[3] = ulp_rand_small_loss_mask16_4_2,
	[4] = ulp_rand_small_loss_mask16_5_2,
	[5] = ulp_rand_small_loss_mask16_6_2,
	[6] = ulp_rand_small_loss_mask16_7_2,
	[7] = ulp_rand_small_loss_mask16_8_2,
	[8] = ulp_rand_small_loss_mask16_9_2,
	[9] = ulp_rand_small_loss_mask16_10_2,
	[10] = ulp_rand_small_loss_mask16_11_2,
	[11] = ulp_rand_small_loss_mask16_12_2,
};

mask_set_t ulp_rand_small_loss_masks16_set[] = {
	 {ulp_rand_small_loss_mask16_1_1, 1, 1},
	 {ulp_rand_small_loss_mask16_2_1, 2, 1},
	 {ulp_rand_small_loss_mask16_3_2, 3, 2},
	 {ulp_rand_small_loss_mask16_4_2, 4, 2},
	 {ulp_rand_small_loss_mask16_5_2, 5, 2},
	 {ulp_rand_small_loss_mask16_6_2, 6, 2},
	 {ulp_rand_small_loss_mask16_7_2, 7, 2},
	 {ulp_rand_small_loss_mask16_8_2, 8, 2},
	 {ulp_rand_small_loss_mask16_9_2, 9, 2},
	 {ulp_rand_small_loss_mask16_10_2, 10, 2},
	 {ulp_rand_small_loss_mask16_11_2, 11, 2},
	 {ulp_rand_small_loss_mask16_12_2, 12, 2}
}; 

mask_set_t ulp_rand_medium_loss_masks16_set[] = {
	 {ulp_rand_medium_loss_mask16_1_1, 1, 1},
	 {ulp_rand_medium_loss_mask16_2_1, 2, 1},
	 {ulp_rand_medium_loss_mask16_3_2, 3, 2},
	 {ulp_rand_medium_loss_mask16_4_2, 4, 2},
	 {ulp_rand_medium_loss_mask16_5_3, 5, 3},
	 {ulp_rand_medium_loss_mask16_6_3, 6, 3},
	 {ulp_rand_medium_loss_mask16_7_4, 7, 4},
	 {ulp_rand_medium_loss_mask16_8_4, 8, 4},
	 {ulp_rand_medium_loss_mask16_9_5, 9, 5},
	 {ulp_rand_medium_loss_mask16_10_5, 10, 5},
	 {ulp_rand_medium_loss_mask16_11_6, 11, 6},
	 {ulp_rand_medium_loss_mask16_12_6, 12, 6}
}; 

mask_set_t ulp_rand_high_loss_masks16_set[] = {
	 {ulp_rand_high_loss_mask16_1_1, 1, 1},
	 {ulp_rand_high_loss_mask16_2_2, 2, 2},
	 {ulp_rand_high_loss_mask16_3_3, 3, 3},
	 {ulp_rand_high_loss_mask16_4_4, 4, 4},
	 {ulp_rand_high_loss_mask16_5_5, 5, 5},
	 {ulp_rand_high_loss_mask16_6_6, 6, 6},
	 {ulp_rand_high_loss_mask16_7_7, 7, 7},
	 {ulp_rand_high_loss_mask16_8_8, 8, 8},
	 {ulp_rand_high_loss_mask16_9_9, 9, 9},
	 {ulp_rand_high_loss_mask16_10_10, 10, 10},
	 {ulp_rand_high_loss_mask16_11_11, 11, 11},
	 {ulp_rand_high_loss_mask16_12_12, 12, 12}
}; 

uint16_t** ulp_rand_loss_masks_sets_list[] = {
	/* 0 - smallest loss, 9 - highest loss */
	[0] = ulp_rand_small_loss_masks16,
	[1] = ulp_rand_small_loss_masks16,
	[2] = ulp_rand_small_loss_masks16,
	[3] = ulp_rand_medium_loss_masks16,
	[4] = ulp_rand_medium_loss_masks16,
	[5] = ulp_rand_medium_loss_masks16,
	[6] = ulp_rand_medium_loss_masks16,
	[7] = ulp_rand_high_loss_masks16,
	[8] = ulp_rand_high_loss_masks16,
	[9] = ulp_rand_high_loss_masks16
};

mask_set_t* ulp_rand_loss_masks_sets_list_set[] = {
	[0] = (mask_set_t*)&ulp_rand_small_loss_masks16_set,
	[1] = (mask_set_t*)&ulp_rand_small_loss_masks16_set,
	[2] = (mask_set_t*)&ulp_rand_small_loss_masks16_set,
	[3] = (mask_set_t*)&ulp_rand_medium_loss_masks16_set,
	[4] = (mask_set_t*)&ulp_rand_medium_loss_masks16_set,
	[5] = (mask_set_t*)&ulp_rand_medium_loss_masks16_set,
	[6] = (mask_set_t*)&ulp_rand_medium_loss_masks16_set,
	[7] = (mask_set_t*)&ulp_rand_high_loss_masks16_set,
	[8] = (mask_set_t*)&ulp_rand_high_loss_masks16_set,
	[9] = (mask_set_t*)&ulp_rand_high_loss_masks16_set
};

uint16_t*** ulp_loss_masks_sets_list;
mask_set_t** ulp_loss_masks_sets_list_set;

const char* ulp_status_2_str[] = {
	[0] = "OK",
    [1] = "FAIL",
	[2] = "RECOVERY_PARTIALLY_SUCCEDED",
	[3] = "RECOVERY_FAILED_DUE_MISSING_FECS", 
	[4] = "RECOVERY_FAILED_DUE_MISSING_RTP", 
	[5] = "RECOVERY_FAILED_DUE_SYSTEM_IO_ERR",
	[6] = ""	
};

SWITCH_DECLARE(void) ulp_set_scheme_masks_sets_type(int loss_type)
{
	if (loss_type == 0) {
		/*random loss*/
		ulp_loss_masks_sets_list_set = &ulp_rand_loss_masks_sets_list_set[0];
		ulp_loss_masks_sets_list = ulp_rand_loss_masks_sets_list;
	} else {
		/*bursty loss*/
		ulp_loss_masks_sets_list_set = &ulp_burst_loss_masks_sets_list_set[0];
		ulp_loss_masks_sets_list = ulp_burst_loss_masks_sets_list;
	}
}

SWITCH_DECLARE(int) ulp_get_scheme_fec_count(int set_idx, int mask_idx)
{
	mask_set_t* mset = NULL;   
	
	if (ulp_loss_masks_sets_list_set && ulp_loss_masks_sets_list_set[set_idx]) {
		mset = ulp_loss_masks_sets_list_set[set_idx];
		if (mset) {
			return mset[mask_idx].fec_packets;
		} else {
			return 0;
		}
	}
	return 0;
}

SWITCH_DECLARE(int) ulp_get_scheme_media_count(int set_idx, int mask_idx)
{
	mask_set_t* mset = NULL;   
	
	if (ulp_loss_masks_sets_list_set && ulp_loss_masks_sets_list_set[set_idx]) {
		mset = ulp_loss_masks_sets_list_set[set_idx];
		if (mset) {
			return mset[mask_idx].fec_packets;
		} else {
			return 0;
		}
	}
	return 0;
}

SWITCH_DECLARE(ulp_status_t) ulp_policy_init(ulp_policy_t *policy,
										uint32_t ssrc,
										ulp_cb_rtp_material cb_rtp_material,
										void* user_data,
										ulp_cb_rtp_by_seq cb_rtp_by_seq,
										ulp_cb_find_fecs_for_seq_on_lvl cb_find_fecs_for_seq_on_lvl,
										ulp_cb_rtp_cover cb_rtp_cover)
{
	if (!policy)
		return ulp_status_fail;
	
	/* All receiver's callbacks must be NULL or !NULL. */
	if (!(cb_rtp_by_seq && cb_find_fecs_for_seq_on_lvl && cb_rtp_cover) && (cb_rtp_by_seq || cb_find_fecs_for_seq_on_lvl || cb_rtp_cover)) {
		return ulp_status_fail;
	}

	memset(policy, 0, sizeof(ulp_policy_t));

	policy->ssrc = ssrc;
	policy->user_data = user_data;

	/* Sender's callback. */	
	policy->cb_rtp_material = cb_rtp_material;
	
	/* Receiver's callbacks. */	
	policy->cb_rtp_by_seq = cb_rtp_by_seq;
	policy->cb_find_fecs_for_seq_on_lvl = cb_find_fecs_for_seq_on_lvl;
	policy->cb_rtp_cover = cb_rtp_cover;

	return ulp_status_ok;
}

SWITCH_DECLARE(void) ulp_policy_deinit(ulp_policy_t *policy)
{
}

SWITCH_DECLARE(size_t) ulp_attempted_recovery_n(ulp_policy_t *policy)
{
	if (!policy)
		return 0;

	return policy->attempted_recovery_n;
}

SWITCH_DECLARE(void) ulp_increment_attempted_recovery_n(ulp_policy_t *policy)
{
	if (!policy)
		return;

	policy->attempted_recovery_n++;
}

SWITCH_DECLARE(size_t) ulp_attempted_missing_fec(ulp_policy_t *policy)
{
	if (!policy)
		return 0;

	return policy->attempted_missing_fec;
}

SWITCH_DECLARE(void) ulp_increment_attempted_missing_fec(ulp_policy_t *policy)
{
	if (!policy)
		return;

	policy->attempted_missing_fec++;
}

SWITCH_DECLARE(size_t) ulp_attempted_missing_media(ulp_policy_t *policy)
{
	if (!policy)
		return 0;

	return policy->attempted_missing_media;
}

SWITCH_DECLARE(void) ulp_increment_attempted_missing_media(ulp_policy_t *policy)
{
	if (!policy)
		return;

	policy->attempted_missing_media++;
}

SWITCH_DECLARE(size_t) ulp_successful_recovery_n(ulp_policy_t *policy)
{
	if (!policy)
		return 0;

	return policy->successful_recovery_n;
}

SWITCH_DECLARE(void) ulp_increment_successful_recovery_n(ulp_policy_t *policy)
{
	if (!policy)
		return;

	policy->successful_recovery_n++;
}

SWITCH_DECLARE(size_t) ulp_successful_partial_recovery_n(ulp_policy_t *policy)
{
	if (!policy)
		return 0;

	return policy->successful_partial_recovery_n;
}

SWITCH_DECLARE(void) ulp_increment_successful_partial_recovery_n(ulp_policy_t *policy)
{
	if (!policy)
		return;

	policy->successful_partial_recovery_n++;
}

SWITCH_DECLARE(uint8_t) ulp_lvl_n(ulp_policy_t *policy)
{
	return policy->lvl_n;
}

SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_n(ulp_policy_t *policy, uint8_t n)
{
	if (n > ULP_LVL_N) {
		return ulp_status_fail;
	}

	policy->lvl_n = n;

	return ulp_status_ok;
}

SWITCH_DECLARE(uint8_t) ulp_lvl_group_size(ulp_policy_t *policy, uint8_t k)
{
	return policy->lvl[k].gk;
}

/* TODO: assert all levels g_k follow the requirement that each must be multiple of previous one. */
SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_group_size(ulp_policy_t *policy, uint8_t k, uint8_t g_k)
{
	if (!policy || (k > ULP_LVL_N - 1) || (g_k > ULP_LVL_GROUP_SIZE_MAX)) {
		return ulp_status_fail;
	}

	policy->lvl[k].gk = g_k;

	return ulp_status_ok;
}

SWITCH_DECLARE(uint16_t) ulp_lvl_protection_len(ulp_policy_t *policy, uint8_t k)
{
	return policy->lvl[k].lk;
}

SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_protection_len(ulp_policy_t *policy, uint8_t k, uint16_t len)
{
	if (!policy || (k > ULP_LVL_N - 1) || (len > ULP_LVL_LEN_MAX)) {
		return ulp_status_fail;
	}

	policy->lvl[k].lk = len;

	return ulp_status_ok;
}

SWITCH_DECLARE(uint16_t) ulp_lvl_protection_mask16(ulp_policy_t *policy, uint8_t k)
{
	return policy->lvl[k].mask16;
}

SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_protection_mask16(ulp_policy_t *policy, uint8_t k, uint16_t mask16)
{
	if (!policy || (k > ULP_LVL_N - 1)) {
		return ulp_status_fail;
	}

	policy->lvl[k].mask16 = mask16;

	return ulp_status_ok;
}

SWITCH_DECLARE(uint32_t) ulp_lvl_protection_mask_cont(ulp_policy_t *policy, uint8_t k)
{
	return policy->lvl[k].mask_cont;
}

SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_protection_mask_cont(ulp_policy_t *policy, uint8_t k, uint32_t mask_cont)
{
	if (!policy || (k > ULP_LVL_N - 1)) {
		return ulp_status_fail;
	}

	policy->lvl[k].mask_cont = mask_cont;

	return ulp_status_ok;
}

SWITCH_DECLARE(uint8_t*) ulp_fec_payload(ulp_fec_lvl_64_info_t *lvl, uint8_t lvl_n, ulp_fec_pkt_t *fec, uint8_t k)
{
	uint8_t i = 0;
	uint16_t lk_cum = 0;

	if (k + 1 > lvl_n) {
		return NULL;
	}

	while (i < k) {
		lk_cum += lvl[i].lk;
		++i;
	}

	return ((uint8_t *) &fec->lvl[0] + (k + 1) * (fec->hdr.l ? 8 : 4) + lk_cum);
}

SWITCH_DECLARE(uint16_t) ulp_fec_last_len(ulp_policy_t *policy)
{
	return policy->fec_last_len;
}

SWITCH_DECLARE(void) ulp_fec_set_last_len(ulp_policy_t *policy, uint16_t fec_len)
{
	policy->fec_last_len = fec_len;
}

SWITCH_DECLARE(uint16_t) ulp_fec_max_len(ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint8_t lvl_n)
{
	uint8_t i = 0;
	uint16_t lk_cum = 0;
	uint8_t lvl_n_parsed = 0;
	ulp_fec_lvl_64_info_t lvl[ULP_LVL_N] = {{0}};
	uint8_t more = 0;
	uint8_t l = 0;


	if (ulp_fec_parse(NULL, fec, sizeof(ulp_fec_hdr_t) + fec_payload_len, lvl, &lvl_n_parsed,  &more, &l) != ulp_status_ok) {
		return 0;
	}

	while (i < ulp_min(lvl_n, lvl_n_parsed)) {
		lk_cum += lvl[i].lk;
		++i;
	}

	return (sizeof(ulp_fec_hdr_t) + i * (l ? 8 : 4) + lk_cum);
}

SWITCH_DECLARE(void) ulp_rtp_set_last_len(ulp_policy_t *policy, uint16_t rtp_len)
{
	policy->rtp_last_len = rtp_len;
}

SWITCH_DECLARE(uint16_t) ulp_rtp_last_len(ulp_policy_t *policy)
{
	return policy->rtp_last_len;
}

SWITCH_DECLARE(void) ulp_rtp_set_last_recover_len(ulp_policy_t *policy, uint16_t rtp_len)
{
	policy->rtp_last_recover_len = rtp_len;
}

SWITCH_DECLARE(uint16_t) ulp_rtp_last_recover_len(ulp_policy_t *policy)
{
	return policy->rtp_last_recover_len;
}

SWITCH_DECLARE(void) ulp_xor(void *x, void *y, size_t len)
{
	uint8_t *xp = x;
	uint8_t *yp = y;

	for (; len--; ++xp, ++yp) {
		*xp ^= *yp;
	}
}

SWITCH_DECLARE(uint8_t) ulp_fec_mask_to_gk(uint64_t mask)
{
	uint8_t gk = 0;
	int i = 48;

	while (mask > 0 && i) {
		
		if (mask & 0x8000000000000000) {
			++gk;
		}

		mask = mask << 1;
		mask &= 0xFFFFFFFFFFFF0000;
		--i;
	}

	return gk;
}

/**
 * Return the @i-th sequence set in the mask. Assume MSB in mask denotes @snb
 * and sequence for the first bit set in the mask is returned for @i == 1.
 */
SWITCH_DECLARE(uint16_t) ulp_mask_2_seq(uint16_t mask16, uint32_t mask_cont, uint16_t snb, uint8_t i)
{
	uint8_t bit = 0;
	uint64_t mask = ((uint64_t) mask16 << 48) | ((uint64_t) mask_cont << 16);
	int j = 48;

	mask &= 0xFFFFFFFFFFFF0000;

	while (i && j) {
		if (mask & 0x8000000000000000) {
			--i;
		}
		
		mask = mask << 1;
		mask &= 0xFFFFFFFFFFFF0000;
		--j;
		++bit;

		/* If not found. */
		if (mask == 0 && i > 0) {
			return 0;
		}
	}

	return snb + bit - 1;
}

/**
 * For ULPFEC:
 *
 * RFC 5109 7.4 FEC Level Header for FEC Packets
 *
 * For a media packet to be protected at level p, it MUST also be
 * protected at level p-1 in any FEC packets.  Please note that the
 * protection level p for a media packet can be in an FEC packet that
 * is different from the one that contains protection level p-1 for
 * the same media packet.
 */
SWITCH_DECLARE(ulp_fec_pkt_t*) ulp_gen_fec(ulp_policy_t *policy, uint8_t fec_n, uint16_t snb)
{
	ulp_fec_pkt_t	*fec, *tmp = NULL;
	uint8_t			lvl_pkt_sz = 4;
	uint16_t		fec_len = 0;
	uint8_t			k = 0;
	ulp_rtp_msg_t	*e = NULL;
	uint8_t			i = 0;
	uint16_t		total_fbs_len = 0;		/* total possible length of the FEC bit string = sum of the lengths of all protection levels */
	uint16_t		k_fbs_len = 0;			/* length of the FEC bit string on a protection level @k */
	uint16_t		k_payload_len_max = 0;	/* length of the longest RTP packet on protection level @k */

	uint16_t *lvl_len = NULL;
	uint16_t *lvl_mask = NULL;
	uint16_t protected_cum = 0;
	uint32_t mask32 = 0;

	uint8_t	pkts_in_buf = 0;
	
	ulp_rtp_msg_t	rtp[ULP_LVL_GROUP_SIZE_MAX];	/* RTP packets needed for recovery. */

	uint8_t use_ulp = 0;


	if (!policy || (fec_n < 1) || (fec_n > ULP_LVL_N)) {
		return NULL;
	}
	
	memset(rtp, 0, sizeof(rtp));
	
	/**
	 * Check if ULP scheme is used.
	 */
	use_ulp = ulp_test_flag(policy, ULP_FL_USE_ULP);
	
	/* Make space for FEC hdr (10 octets) plus each level's FEC level header. */
	
	fec_len = sizeof(ulp_fec_hdr_t);

	k = 0;

	while (k < fec_n) {
		
		/* Cannot protect packet if RTP sequences are not known or number of packets is 0. */
		if ((policy->lvl[k].mask16 == 0 && policy->lvl[k].mask_cont == 0) || (policy->lvl[k].gk == 0)) {
			return NULL;
		}

		if (policy->lvl[k].mask_cont > 0) {
			lvl_pkt_sz = 8;
		}

		total_fbs_len += policy->lvl[k].lk;

		++k;
	}

	/* Now set the FEC packet size, accounting also for all level headers. This is maximum possible size of the FEC, which may be resized at the end of this function if any of the protection level's length is more than longest RTP on that level. */
	fec_len += total_fbs_len + fec_n * lvl_pkt_sz;

	fec = calloc(fec_len, 1);
	if (!fec) {
		goto ulp_gen_fec_err;
	}

	fec->hdr.e = 0;
	fec->hdr.l = (lvl_pkt_sz == 4 ? 0 : 1);


	/* For all protection levels. */

	k = 0;

	while (k < fec_n) {
		
		/* Get RTP packets needed by FEC at this protection level. */
		pkts_in_buf = policy->cb_rtp_material(policy, snb, k, rtp);

		if (pkts_in_buf != policy->lvl[k].gk) {
			goto ulp_gen_fec_err; /* Not possible to generate FEC on this level */
		}

		/* Find length of the longest packet on level @k. */
		k_payload_len_max = 0;
		
		for (i = 0; i < policy->lvl[k].gk; ++i) {

			e = &rtp[i];

			if (e->rtp == NULL) {
				goto ulp_gen_fec_err;
			}
			
			if (e->payload_original_len > k_payload_len_max) {
				k_payload_len_max = e->payload_original_len;
			}
		}

		/* Protected RTP on that level. */
		k_fbs_len = (use_ulp ? (k_payload_len_max > protected_cum ? ulp_min(k_payload_len_max - protected_cum, policy->lvl[k].lk) : 0) : ulp_min(k_payload_len_max, policy->lvl[k].lk));

		/**
		 * Prepare packets for FEC payload generation.
		 * All RTP packets MUST be padded up to the length of the FEC bit string length of that level.
		 */
		ulp_rtp_pad(rtp, pkts_in_buf, (use_ulp ? protected_cum + k_fbs_len : k_fbs_len));

		/* For ULP scheme L0 is special, as only for this level RTP headers are protected.
		 * TODO: Not sure about Chromium's version of FEC, which uses many protection levels
		 * to protect same bytes in packet (not ULP). Since each mask may use different set of packets
		 * which mask (and set of packets) were used for the header? Assume it is the same, i.e level 0,
		 * but this needs to be confirmed.
		 *
		 * NOTE: confirmed Chromium doesn't stack multiple such Non-ulp headers in a single FEC - only L0 is sent in a single FEC.
		 **/
		if (k == 0) {

			for (i = 0; i < policy->lvl[0].gk; ++i) {

				e = &rtp[i];
				
				/* Indirectly XOR length recovery field. */
				if (i == 0) {
					memcpy(&fec->hdr, &e->rtp->hdr, 8);
					fec->hdr.len_r = htons(e->payload_original_len);		/* payload len must be in the network byte order */
				} else {
					ulp_xor(&fec->hdr, &e->rtp->hdr, 8);					/* XOR only first 8 octets. Do not XOR length recovery field. */
					fec->hdr.len_r ^= htons(e->payload_original_len);		/* payload len must be in the network byte order */
				}
			}

			/* This is excluded from XOR. */
			fec->hdr.e = 0;													/* reserved for future use */
			fec->hdr.l = (lvl_pkt_sz == 4 ? 0 : 1);
			fec->hdr.snb = htons(snb);

			/* FEC Header is almost ready now (SN Base will be added later). */

			lvl_len = (uint16_t *)((uint8_t *) fec + sizeof(ulp_fec_hdr_t));
			lvl_mask = lvl_len + 1;	/* 16 bits from lvl_len */
		}

		/**
		 * FEC Level header.
		 */	
		*lvl_len = k_fbs_len;												/* changed to the network order later */

		*lvl_mask = htons(policy->lvl[k].mask16);
		if (fec->hdr.l) {
			mask32 = htonl(policy->lvl[k].mask_cont);
			memcpy(lvl_mask + 1, &mask32, 4);
		}
		
		/**
		 * FEC Level payload.
		 * XOR protected bit strings for all packets protected by this level.
		 */
		for (i = 0; i < policy->lvl[k].gk; ++i) {
			
			e = &rtp[i];

			if (i == 0) {
				memcpy((char *) lvl_len + lvl_pkt_sz, (char *) e->rtp->b + (use_ulp ? protected_cum : 0), *lvl_len);
			} else {
				ulp_xor((char *) lvl_len + lvl_pkt_sz, (char *) e->rtp->b + (use_ulp ? protected_cum : 0), *lvl_len);	/* XOR the FEC payload protected bit strings */
			}
		}
	
		/* For ULP, accumulate total protected bytes. */	
		protected_cum += *lvl_len;
	
		/* Finalise header. */	
		*lvl_len = htons(*lvl_len);

		/* Set pointers to next level's length and mask. */
		lvl_len = (uint16_t *)((char *) lvl_len + lvl_pkt_sz + k_fbs_len);
		lvl_mask = lvl_len + 1; /* 16 bits from lvl_len */
	
		ulp_rtp_free(rtp, pkts_in_buf);

		++k;
	}

	/* Actual length of the FEC packet. */
	total_fbs_len = sizeof(ulp_fec_hdr_t) + protected_cum +	fec_n * lvl_pkt_sz;
	
	if (total_fbs_len != fec_len) {
		
		fec_len = total_fbs_len;

		tmp = fec;
		fec = realloc(tmp, fec_len);
		if (!fec) {
			free(tmp);
			goto ulp_gen_fec_err;
		}
	}

	ulp_fec_set_last_len(policy, fec_len);

	return fec;

ulp_gen_fec_err:

	ulp_rtp_free(rtp, pkts_in_buf);
	free(fec);
	return NULL;
}

/**
 * All arguments in host order.
 */
SWITCH_DECLARE(uint8_t) ulp_seq_in_mask(uint16_t snb, uint16_t seq, uint16_t mask16, uint32_t mask_cont)
{
	uint8_t l = mask_cont > 0 ? 1: 0;

	/* Check if @seq is protected. */
		 if (((seq <= snb + 15) && (mask16 & (0x8000u >> (seq - snb)))) || (l && (seq - snb > 15) && (mask_cont & (0x80000000u >> (seq - snb - 16))))) {
			return 1;
		}
	return 0;
}

/**
 * Return 1 if the FEC packet protects @seq on the protection level @k, otherwise return 0.
 * This function parses FEC packet and checks if it contains required level of protection,
 * and if protection mask of that level includes @seq.
 *
 * @seq in host byte order.
 */
SWITCH_DECLARE(uint8_t) ulp_seq_protected_by_fec_on_lvl(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint16_t seq, uint8_t k)
{
	uint16_t mask16 = 0;
	uint32_t mask_cont = 0;

	ulp_fec_lvl_64_info_t lvl[ULP_LVL_N] = {{0}};
	uint8_t lvl_n = ULP_LVL_N;
	uint8_t more = 0;
	uint8_t l = 0;

	/* 1. Check if FEC's @snb is not higher than @seq. If it is above @seq, this FEC cannot protect @seq on any level of protection,
	 * as the protection mask describes packets with linearly increasing sequences starting from @snb. */

	if (ntohs(fec->hdr.snb) > seq) {
		return 0;
	}

	/* 2. Understand the levels of protection in this FEC. */

	if (ulp_fec_parse(policy, fec, sizeof(ulp_fec_hdr_t) + fec_payload_len, lvl, &lvl_n,  &more, &l) != ulp_status_ok) {
		return 0;
	}
	
	/* 3. Check if FEC contains required level of protection. */ 

	if (lvl_n < k + 1) {
		return 0;
	}

	/* 4. Get the mask and the number of packets protected on that level. */
	
	mask16 = lvl[k].mask16;				/* levels contain host ordered info */
	mask_cont = lvl[k].mask_cont;

	/* 5. Check if @seq is protected. */

	return ulp_seq_in_mask(ntohs(fec->hdr.snb), seq, mask16, mask_cont);
}

/* Pad packets shorter than @pad to @pad length. */
SWITCH_DECLARE(ulp_status_t) ulp_rtp_pad(ulp_rtp_msg_t *rtp, uint8_t pkts, uint16_t pad)
{
	ulp_rtp_msg_t	*e = NULL;							/* iterator over buffer */
	int i = 0;

	for (i = 0; i < pkts; ++i) {
		e = &rtp[i];
		
			if (pad && (pad > e->payload_len)) {
				/* Padding needed. */
				e->rtp = realloc(e->rtp, sizeof(ulp_rtp_hdr_t) + pad);
				if (!e->rtp) {
					return ulp_status_fail;
				}
				memset((uint8_t*)e->rtp + e->payload_len + sizeof(ulp_rtp_hdr_t), 0, pad - e->payload_len);
				e->payload_len = pad;
			}
	}

	return ulp_status_ok;
}

SWITCH_DECLARE(void) ulp_rtp_free(ulp_rtp_msg_t *rtp, uint8_t pkts)
{
	ulp_rtp_msg_t	*e = NULL;							/* iterator over buffer */
	int i = 0;

	for (i = 0; i < pkts; ++i) {
		e = &rtp[i];
		
			if (e->rtp) {
				free(e->rtp);
				memset(e, 0, sizeof(ulp_rtp_msg_t));
			}
	}
}

SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_recover_hdr(ulp_fec_pkt_t *fec, uint16_t seq, uint8_t gk, ulp_rtp_msg_t *rtp, uint16_t *len_r)
{
	ulp_rtp_pkt_t *new_rtp = NULL;
	ulp_rtp_hdr_t *new_hdr = NULL;
	uint8_t rtp_n = 0;

	/* Allocate space for the RTP header. */
	new_hdr = calloc(1, sizeof(ulp_rtp_hdr_t));
	if (!new_hdr) {
		return NULL;
	}
	new_rtp = (ulp_rtp_pkt_t *)new_hdr;

	/* RTP header recovery. Protected bit strings are first 8 octets from RTP and FEC headers. */
	memcpy(new_rtp, &fec->hdr, 8);
	*len_r = ntohs(fec->hdr.len_r);

	for (rtp_n = 0; rtp_n + 1 < gk; ++rtp_n) {
		ulp_xor(new_rtp, rtp[rtp_n].rtp, 8);
		*len_r ^= rtp[rtp_n].payload_original_len;
	}

	/* Resize RTP packet to the to-recover size. */
	new_rtp = realloc(new_rtp, sizeof(ulp_rtp_hdr_t) + *len_r);
	memset((char*)new_rtp + sizeof(ulp_rtp_hdr_t), 0, *len_r); 

	new_rtp->hdr.v = 2;
	new_rtp->hdr.seq = htons(seq);
	/** Note: SSRC skipped. */

	return new_rtp;
}

/**
 * Return best of the recovered packets (by recovered size) and free all the other.
 */
static ulp_rtp_pkt_t* ulp_recovery_best(ulp_recovered_rtp_t *new_rtps, uint16_t new_rtp_n, uint16_t *recovered_len)
{
	ulp_rtp_pkt_t *best = NULL;
	int i = 0;
	uint16_t limit = ulp_min(ULP_RECOVERED_RTP_MAX_N, new_rtp_n);
	
	
	*recovered_len = 0;

	while (i < limit) {

		if (new_rtps[i].rtp) {

			if (new_rtps[i].rtp_len >= *recovered_len) {

				if (best) {
					free(best);
				}

				best = new_rtps[i].rtp;
				*recovered_len = new_rtps[i].rtp_len;

			} else {
				free(new_rtps[i].rtp);
				new_rtps[i].rtp = NULL;
			}
		}

		++i;
	}

	return best;
}

/* TODO Consider adding recovered RTP to the circular buffer - it will allow for some of the recoveries which otherwise would fail.
 * Consider: G(0) = 2, G(1) = 4.
 * 1. put D in the buf, C is missing
 * 2. put FEC1 in buf (contains L0 for C and D, and maybe L1 for A, B, C, D)
 * 3. Recover C on L0 (using FEC1 and D) and maybe L1 (using FEC1 and A, B, D if all in buf)
 * 4. put E in buf
 * 5. put FEC2 in buf
 * 6. F is missing, try to recover: recovery possible only on L0 (using E and FEC 2), L1 not possible (need FEC2 C, D, E) as C is missing.
 * If C was put in the buf after recovery, and if it was recovered on L1, then F recovery on L1 would succeed.
 */
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_recover(ulp_policy_t *policy, uint16_t seq, ulp_status_t *err, uint8_t *packet_valid, uint8_t(*cb_validate_recovery)(ulp_rtp_pkt_t *new_rtp, ulp_rtp_msg_t *rtp, uint32_t rtp_n))
{
	ulp_rtp_pkt_t	*new_rtp = NULL, *tmp = NULL;
	ulp_recovered_rtp_t	new_rtps[ULP_RECOVERED_RTP_MAX_N];					/* Recovered versions of the missing RTP packet. */
	uint16_t			new_rtp_n = 0;										/* Number of recovered versions of packet. */
	uint8_t rtp_n = 0;
	uint8_t fec_n = 0, i = 0;

	ulp_fec_pkt_t	*fec = NULL;					/* FEC packet needed for recovery of the RTP packet. */
	uint16_t		fec_payload_len = 0;
	uint8_t			gk = 0;							/* Group size of level k, G(k). */
	uint16_t		lk = 0, lk_max = 0;				/* protection length in octets */
	ulp_rtp_msg_t	rtp[ULP_LVL_GROUP_SIZE_MAX];	/* RTP packets needed for recovery. */
	ulp_rtp_msg_t	fecs[ULP_LVL_GROUP_SIZE_MAX];	/* FECs needed for recovery. */
	uint8_t			k = 0;							/* Protection level. */				
	uint16_t		recovered = 0;					/* Maximum recovered length of the packet. */
	uint16_t		prev_len_r = 0, len_r = 0;						/* length of the missing RTP packet, ideally recovered in total */
	uint16_t		b2r = 0;						/* total packet payload length to recover */
	uint16_t		r = 0;							/* lentgh recovered in one go of the loop */
	
	ulp_fec_lvl_64_info_t lvl[ULP_LVL_N] = {{0}};
	uint8_t lvl_n = ULP_LVL_N;
	uint8_t more = 0;
	uint8_t l = 0;

	uint8_t use_ulp = 0;
	uint8_t valid = 1;


	*err = ulp_status_fail;

	if (!policy || !policy->cb_rtp_by_seq || !policy->cb_find_fecs_for_seq_on_lvl || !policy->cb_rtp_cover) {
		return NULL;
	}

	ulp_increment_attempted_recovery_n(policy);

	/* JB's node->packet.body contains the RTP header too, we also keep the header in node->packet.header */
	new_rtp = policy->cb_rtp_by_seq(policy, seq, &recovered);

	if (new_rtp) {
		ulp_rtp_set_last_len(policy, recovered);
		ulp_rtp_set_last_recover_len(policy, recovered);
		*err = ulp_status_ok;
		*packet_valid = valid;
		ulp_increment_successful_recovery_n(policy);

		return new_rtp;
	}

	/**
	 * Check if ULP scheme is used.
	 */
	use_ulp = ulp_test_flag(policy, ULP_FL_USE_ULP);

	memset(rtp, 0, sizeof(rtp));
	memset(fecs, 0, sizeof(fecs));
	memset(new_rtps, 0, sizeof(new_rtps));


	/* Determine if set of packets needed for recovery is available in the buffer. */
	/* There may be many FEC packets associated with same @seq. */

	/*
	 * Search through all FECs for these which can be used to recover this RTP packet.
	 */

	/**
	 * For all protection levels.
	 */

	for (k = 0; k < ULP_LVL_N; ++k) {

		memset(fecs, 0, sizeof(fecs));

		/**
		 * Find all FECs protecting @seq on level @k. Each FEC returned in the @fecs table can be used to recover @seq,
		 * because FEC contains @seq in a bitmask of level @k.
		 */

		fec_n = policy->cb_find_fecs_for_seq_on_lvl(policy, seq, k, fecs);

		if (fec_n < 1) {
			
			if (k == 0) {
				
				*err = ulp_status_recovery_no_fecs;
				ulp_increment_attempted_missing_fec(policy);
				goto ulp_recover_err;											/* Recovery not possible, cannot reconstruct the header. */

			} else {

				continue;
			}
		}

		/**
		 * For all FECs on level @k.
		 */

		/* For ULP each level is separate recovery, for non-ulp each level is just as if it was L0. */
		if (use_ulp) {
			lk_max = 0;
		}

		for (i = 0; i < fec_n; ++i) {

			fec = (ulp_fec_pkt_t *) fecs[i].rtp;
			fec_payload_len = fecs[i].payload_len;
	

			/* Understand the levels of protection in this FEC. */

			lvl_n = ULP_LVL_N;
			memset(lvl, 0, sizeof(lvl));

			if (ulp_fec_parse(policy, fec, sizeof(ulp_fec_hdr_t) + fec_payload_len, lvl, &lvl_n,  &more, &l) != ulp_status_ok) {
				return 0;
			}

			gk = lvl[k].gk;
			lk = lvl[k].lk;

			switch_assert(!((lvl[k].mask16 != 0 || lvl[k].mask_cont != 0) && gk == 0));
			/**
			 * For non-ulp, do not attempt recovery if the possible recovery length is less what we've already recovered.
			 */
			if (!use_ulp && (lk < lk_max)) {
				continue;
			}
			
			/* Validate that this FEC packet is correct FEC packet for { @snb, @seq, @k }. */
			if (!ulp_seq_in_mask(ntohs(fec->hdr.snb), seq, lvl[k].mask16, lvl[k].mask_cont)) {
				continue;
			}
			/* We must have G(k) - 1 RTP packets to be able to recover payload on level k. */
			/* Receives copies of packets. */
			rtp_n = policy->cb_rtp_cover(policy, fec, fec_payload_len, k, rtp);

			if (rtp_n + 1 < gk) {

				ulp_rtp_free(rtp, rtp_n);
				
				if (k == 0) {
					
					/**
					 * Recovery not possible, cannot reconstruct the header.
					 */
					
					*err = ulp_status_recovery_more_rtp;
					ulp_increment_attempted_missing_media(policy);
					goto ulp_recover_err;
				
				} else {

					continue;
				}
			}

			if ((lk > SWITCH_RTP_MAX_BUF_LEN - sizeof(ulp_rtp_hdr_t) - 1) || 
				(rtp[rtp_n].payload_original_len > SWITCH_RTP_MAX_BUF_LEN - sizeof(ulp_rtp_hdr_t) - 1)) {
				goto ulp_recover_err;
			}
			/* Reallocs and pads all packets up to current recovery length @lk. Valid pointers in @rtp table. */
			ulp_rtp_pad(rtp, rtp_n, (use_ulp ? recovered + lk : ulp_max(lk, rtp[rtp_n].payload_original_len)));

			if (k == 0) {

				new_rtp = ulp_recover_hdr(fec, seq, gk, rtp, &len_r);
				if (!new_rtp) {
					*err = ulp_status_recovery_err_mem;
					goto ulp_recover_err;
				}
				
				b2r = len_r;
				if (b2r > SWITCH_RTP_MAX_BUF_LEN - sizeof(ulp_rtp_hdr_t) - 1) {
					goto ulp_recover_err;
				}
			} else {
				
				if (!use_ulp) {
					if (!new_rtp) {
						prev_len_r = len_r;
						new_rtp = calloc(1, sizeof(ulp_rtp_hdr_t) + len_r);
						if (!new_rtp) {
							*err = ulp_status_recovery_err_mem;
							goto ulp_recover_err;
						}
					} else if (len_r > prev_len_r) {
						prev_len_r = len_r;
						tmp = new_rtp;
						new_rtp = realloc(tmp, sizeof(ulp_rtp_hdr_t) + len_r);
						if (!new_rtp) {
							free(tmp);
							*err = ulp_status_recovery_err_mem;
							goto ulp_recover_err;
						}
					}
					if (!new_rtps[0].rtp) {
						goto ulp_recover_err;
					}
					memcpy(new_rtp, new_rtps[0].rtp, 12);
				}
			}

			/* Payload recovery. */
			r = ulp_min(b2r, lk);
			memcpy((char *) new_rtp + sizeof(ulp_rtp_hdr_t) + (use_ulp ? recovered : 0), ulp_fec_payload(lvl, lvl_n, fec, k), r);

			for (rtp_n = 0; rtp_n + 1 < gk; ++rtp_n) {
				ulp_xor((char *) new_rtp + sizeof(ulp_rtp_hdr_t) + (use_ulp ? recovered : 0), (char *) rtp[rtp_n].rtp + sizeof(ulp_rtp_hdr_t) + (use_ulp ? recovered : 0), r);  
			}

			if (cb_validate_recovery) {
				valid = cb_validate_recovery(new_rtp, rtp, gk - 1);
			}

			ulp_rtp_free(rtp, rtp_n);

			if (use_ulp) {
				recovered += r;
				b2r -= r;
			} else {
				recovered = r;
			}

			if (!use_ulp) {
				new_rtps[new_rtp_n].rtp = new_rtp;
				new_rtps[new_rtp_n].rtp_len = recovered;
				new_rtps[new_rtp_n].valid = valid;
				++new_rtp_n;
			}
			
			if (lk > lk_max) {
				lk_max = lk;
			}

			if (recovered == len_r) {
				break;
			}
		}
		
		ulp_rtp_free(fecs, fec_n);
	}

	if (!use_ulp) {
		new_rtp = ulp_recovery_best(new_rtps, new_rtp_n, &recovered);
	}
	
	ulp_rtp_set_last_len(policy, sizeof(ulp_rtp_hdr_t) + len_r);
	ulp_rtp_set_last_recover_len(policy, sizeof(ulp_rtp_hdr_t) + recovered);

	/**
	 * If only partial recovery was possible signal that, otherwise complete recovery succeeded.
	 */
	if (recovered < len_r) {
		*err = ulp_status_recovery_partial;
		ulp_increment_successful_partial_recovery_n(policy);
		ulp_increment_successful_recovery_n(policy);
	} else {
		*err = ulp_status_ok;
		ulp_increment_successful_recovery_n(policy);
	}

	*packet_valid = valid;

	return new_rtp;

ulp_recover_err:

	ulp_rtp_free(rtp, rtp_n);
	ulp_rtp_free(fecs, fec_n);
	if (new_rtp) {
		free(new_rtp);
	}
	return NULL;
}

SWITCH_DECLARE(void*) ulp_get_user_data(ulp_policy_t *policy)
{
	return policy ? policy->user_data : NULL;
}

SWITCH_DECLARE(void) ulp_set_user_data(ulp_policy_t *policy, void* user_data)
{
	policy->user_data = user_data;
}


/* RED */

SWITCH_DECLARE(void) ulp_red_encoding_block_set_ts_off_len(uint8_t *ts_off_len, uint16_t ts, uint16_t len)
{
	memset(ts_off_len, 0, 3);
	ts_off_len[0] = (ts >> 6) & 0xFFu;
   	ts_off_len[1] =  ((ts << 2) & 0xFCu);
   	ts_off_len[1] |=  ((len >> 8) & 0x03u);
   	ts_off_len[2] =  ((len) & 0xFFu);
}

SWITCH_DECLARE(ulp_rtp_hdr_t*) ulp_gen_red_hdr(ulp_rtp_hdr_t *rtp, uint8_t red_pt, void *inline_where)
{
	ulp_rtp_hdr_t *r = inline_where;

	if (!rtp) {
		return NULL; /* Graceful crash. */
	}

	if (r == NULL) {
		r = malloc(sizeof(ulp_rtp_hdr_t));
		if (r == NULL) {
			return NULL;
		}
	}

	memcpy(r, rtp, sizeof(ulp_rtp_hdr_t));
	r->m = 0;								/* RFC 5109: 10.3
											   In the payload format for redundant coding as specified by RFC 2198,
											   the marker bit is lost as soon as the primary coding is carried in the RED packets.
											   So the marker bit cannot be recovered regardless of whether or not the FEC is used.

											  RFC 2198: 4  Limitations
											  The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
											  (containing this marker) is lost, the marker is lost. */
	r->pt = red_pt;

	return r;
}

/* Note: supports single Redundant Encoding Block.
 * TODO: add support for multiple blocks of redundant encoding. */ 
SWITCH_DECLARE(uint16_t) ulp_red_data_len(uint16_t rtp_len, uint16_t redundancy_len)
{
	uint16_t len = 0;
	len += 1 + rtp_len;																					/* Primary Encoding Block Header is 1 octet long. */
	len += (redundancy_len == 0 ? 0 : redundancy_len + sizeof(ulp_red_encoding_block_hdr_t));			/* Redundant Encoding Block Header is complete 4 octets long header. */

	return len;
}

SWITCH_DECLARE(void*) ulp_gen_red_data(ulp_rtp_hdr_t *rtp, uint16_t rtp_len, uint8_t primary_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_data_len, void *inline_where)
{
	uint8_t *rd = inline_where;
	uint16_t len = ulp_red_data_len(rtp_len, redundancy_len);
	ulp_red_encoding_block_hdr_t *en = NULL;


	*total_data_len = 0;

	if (!rtp) {
		return NULL; /* Graceful crash. */
	}

	if (rd == NULL) {
		rd = malloc(len);
		if (rd == NULL) {
			return NULL;
		}
		memset(rd, 0, len);
		inline_where = rd;
	}

	en = (ulp_red_encoding_block_hdr_t *) rd;

	if (redundancy_len != 0) {
		en->f = 1;																/* This is Redundant Encoding Block Header. */
		en->pt = redundancy_pt;													/* Redundancy's payload type, for ULPFEC this is ULP_PT_ULPFEC. */

		ulp_red_encoding_block_set_ts_off_len(en->ts_off_len, red_ts_off_diff, redundancy_len);		/* Produce TS and LEN fields in Redundant Encoding Block Header. This represents diff offset from primary media packet. */

		rd = (uint8_t *) (en + 1);
		memcpy((void *) rd, redundancy, redundancy_len);
		rd += redundancy_len; 
	}

	*rd = 0x7F & primary_pt;													/* This is Primary Encoding Block Header. It contains only F bit and PT.
																				   F is 0 to signal primary encoding, PT is payload type of the RTP media packet.
																				   TS and LEN can be deducted from total packet length and RED RTP header. */
	rd++;
	memcpy(rd, (char* )rtp + sizeof(ulp_rtp_hdr_t), rtp_len);
	*total_data_len = len;
	

	return inline_where;
}

SWITCH_DECLARE(void*) ulp_gen_red_pkt(ulp_rtp_hdr_t *rtp, uint16_t rtp_len, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void* inline_where)
{
	uint8_t *rd = inline_where;
	uint16_t len = sizeof(ulp_rtp_hdr_t) + ulp_red_data_len(rtp_len, redundancy_len);


	*total_len = 0;

	if (!rtp) {
		return NULL; /* Graceful crash. */
	}

	if (rd == NULL) {
		rd = malloc(len);
		if (rd == NULL) {
			return NULL;
		}
		inline_where = rd;
	}

	if (!ulp_gen_red_hdr(rtp, red_pt, rd)) {									/* RED RTP header. */
		return NULL;
	}

	/* Optional Redundant Encoding Block (encoding hdr + FEC) + Primary Encoding
	 * Block (encoding header + RTP media packet). */
	if (!ulp_gen_red_data(rtp, rtp_len, rtp->pt, redundancy, redundancy_len, redundancy_pt, red_ts_off_diff, total_len, rd + sizeof(ulp_rtp_hdr_t))) {
		if (!inline_where) {
			free(rd);
		}
		return NULL;
	}

	*total_len += sizeof(ulp_rtp_hdr_t);

	return inline_where;
}

SWITCH_DECLARE(ulp_rtp_hdr_t*) ulp_red_depaketize(ulp_rtp_hdr_t *rtp, uint16_t *rtp_len, void **redundancy, uint16_t *redundancy_len, uint16_t *red_ts_off_diff, void* inline_where)
{
	uint8_t *rd = NULL;
	ulp_red_encoding_block_hdr_t *en = NULL;
	uint16_t red_fec_ts_diff_offset = 0;
	uint16_t red_fec_len = 0;

	char *prim_encoding = NULL;
	uint8_t prim_encoding_pt = 0;
	uint16_t prim_encoding_len = 0;
	ulp_rtp_hdr_t *rtp_out = NULL;

	*redundancy = NULL;
	*red_ts_off_diff = 0;
	*redundancy_len = 0;

	if (!rtp || (*rtp_len < 1)) {												/* There must be at least Primary Encoding Block Header in the RED payload. */
		return NULL;
	}

	rd = (uint8_t *) (rtp + 1);


	/* Packet validation. */

	if (*rd & 0x80) {															/* Check if this is Redundant Encoding header. */
		en = (void *) rd;														/* Indicate Redundancy presence in the packet. */

		red_fec_ts_diff_offset = ulp_red_encoding_block_get_ts(en->ts_off_len);
		red_fec_len = ulp_red_encoding_block_get_len(en->ts_off_len);

		rd = (uint8_t *) en + sizeof(ulp_red_encoding_block_hdr_t) + red_fec_len;	/* Move on to the Primary Encoding Block Header. */
	}

	if (*rd & 0x80) {
		return NULL;															/* We do not support multiple redundant encodings. TODO: add support for multiple encodings. */
	}

	prim_encoding_pt = *rd & 0x7f;
	prim_encoding = (char *) rd + 1;

	if (sizeof(ulp_red_encoding_block_hdr_t) + red_fec_len >= *rtp_len) {
		return NULL;
	}
	prim_encoding_len = *rtp_len - ((en != NULL ? sizeof(ulp_red_encoding_block_hdr_t) + red_fec_len : 0) + 1);

	/* 
	 * RFC 5109 10.3
	 *
	 * The conversion from RED packets to virtual RTP packets is simply done by
	 * (1) removing any RED block headers and redundant coding data,
	 * and (2) replacing the PT in the RTP header with the PT of the primary coding.
	 */

	if (en) {																	/* Check if this is Redundant Encoding header. */

		if (red_fec_len > 0) {

			*redundancy = malloc(red_fec_len);
			if (! *redundancy) {
				return NULL;
			}
		}
	}

	if (!inline_where) {
		
		inline_where = malloc(prim_encoding_len + sizeof(ulp_rtp_hdr_t));
		if (!inline_where) {
			if (*redundancy) {
				free(*redundancy);
			}
			return NULL;
		}
	}

	if (en) {																	/* Check if this is Redundant Encoding header. */

		if (red_fec_len > 0) {
			memcpy(*redundancy, (void *) (en + 1), red_fec_len);
		}

		*red_ts_off_diff = red_fec_ts_diff_offset;
		*redundancy_len = red_fec_len;
	}

	memcpy(inline_where, rtp, sizeof(ulp_rtp_hdr_t));
	memcpy((char *) inline_where + sizeof(ulp_rtp_hdr_t), prim_encoding, prim_encoding_len);
	rtp_out = inline_where;
	rtp_out->pt = prim_encoding_pt;
	*rtp_len = prim_encoding_len;

	return rtp_out;
}

SWITCH_DECLARE(ulp_status_t) ulp_fec_parse(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_len, ulp_fec_lvl_64_info_t *lvl, uint8_t *lvl_n,  uint8_t *more, uint8_t *l)
{
	uint8_t n = 0;					/* number of protection levels found in this FEC packet */

	char *p = (char *) fec;			/* parsing pointer */
	uint16_t bytes_parsed = 0;		/* bytes already parsed */
	uint16_t lk = 0;				/* protection length of the protection level */
	uint8_t gk = 0;					/* group size of the protection level */
	
	ulp_fec_lvl_32_t *l32 = NULL;
	ulp_fec_lvl_64_t *l64 = NULL;


	if (!fec || fec_len < sizeof(ulp_fec_hdr_t) || !lvl || !lvl_n || !more || !l || *lvl_n < 1) {
		return ulp_status_fail;
	}

	*l = fec->hdr.l;

	p += sizeof(ulp_fec_hdr_t);
	bytes_parsed += sizeof(ulp_fec_hdr_t);
	
	if (!(*l)) {
		l32 = (ulp_fec_lvl_32_t *) p;
	}

	while ((n < *lvl_n) && (bytes_parsed < fec_len)) {
		
		/* parse protection level header */
		if (l32) {
			l32 = (ulp_fec_lvl_32_t *) p;

			if (policy && ulp_test_flag(policy, ULP_FL_REDUNDANT_AS_PRIMARY)) {

				/* Chromium sends Redundant Encoding as Primary. */
				lk = ntohs(l32->lk);

			} else {
				lk = ntohs(l32->lk);
			}

			lvl[n].mask16 = ntohs(l32->mask16);
			gk = ulp_fec_mask_to_gk((((uint64_t)lvl[n].mask16) << 48));
		} else {
			l64 = (ulp_fec_lvl_64_t *) p;
			if (policy && ulp_test_flag(policy, ULP_FL_REDUNDANT_AS_PRIMARY)) {
				
				/* Chromium sends Redundant Encoding as Primary. */
				lk = ntohs(l64->lk);

			} else {
				lk = ntohs(l64->lk);
			}
			lvl[n].mask16 = ntohs(l64->mask16);
			lvl[n].mask_cont = ntohl(l64->mask_cont);
			gk = ulp_fec_mask_to_gk((((uint64_t) lvl[n].mask16) << 48) | ((uint64_t) lvl[n].mask_cont << 16));
		}

		lvl[n].lk = lk;
		lvl[n].gk = gk;
		
		/* move to the next level header */
		p += lk + (*l ? 8 : 4);
		bytes_parsed += lk + (*l ? 8 : 4);
		++n;
	}

	*lvl_n = n;
	*more = 0;
	if (bytes_parsed < fec_len) {
		*more = 1;
	}

	if (policy) {
		
		policy->lvl_n = n;
		memcpy(policy->lvl, lvl, sizeof(lvl[0]) * ULP_LVL_N);
		policy->more = 0;

		if (bytes_parsed < fec_len) {
			policy->more = 1;
		}

		policy->l = *l;
	}	

	return ulp_status_ok;
}

SWITCH_DECLARE(uint32_t) ulp_binomial(uint8_t n, uint8_t k)
{
	if (k >= n || k == 0) {
		return 1;
	}

	return n * ulp_binomial(n - 1, k) / (n - k);
}

SWITCH_DECLARE(double) ulp_opt_fec_scheme(uint8_t loss_per_255, uint8_t max_overhead_per_255, uint8_t *fec_n_opt, uint8_t *rtp_n_opt)
{
	uint8_t fec_n = 1;
	uint8_t rtp_n = 1;

	double p = (double) loss_per_255 / 255;	/* Bernoulli probability of single packet loss. */
	double p_recovery = 0;					/* Probability of successful recovery of a single packet. */
	double p_opt = 0;						/* Optimum probability of successful recovery. */
	uint8_t f_n_opt = 0;					/* Optimum number of FECs (and optimum number of packets lost in burst covered). */
	uint8_t r_n_opt = 0;					/* Optimum number of media. */

	uint8_t expected_loss = 0;				/* Expected number of packets lost in packets sent by FEC scheme in a single transaction (rtp_n). */
	uint8_t overhead = 0;					/* 1 if scheme overhead is above limit. */

	uint8_t i = 0;


	/**
	 * If loss is 0 then return 0.
	 */
	if (!loss_per_255) {
		p_opt = 1;
		f_n_opt = 0;
		r_n_opt = 0;
		goto ulp_opt_fec_scheme_done;
	}

	/**
	 * Check allowed overhead.
	 */
	if (max_overhead_per_255 && (255 * 1 > (ULP_MAX_FEC_IN_SCHEME + 1) * max_overhead_per_255)) {
		goto ulp_opt_fec_scheme_done;
	}

	/**
	 * If loss is less than 1/13 start from default.
	 */
	if (255 * 1 > (ULP_MAX_FEC_IN_SCHEME + 1) * loss_per_255) {
		p_opt = (double) 1 / 13;
		f_n_opt = 1;
		r_n_opt = 12;
	}

	/**
	 * For each media number of packets.
	 */

	while (rtp_n <= ULP_MAX_FEC_IN_SCHEME) {

		expected_loss = ulp_max(1, 0.5 + rtp_n * p);

		/** 
		 * For each FEC number of packets per given number of media RTP packets.
		 */
		
		fec_n = 1;

		while (fec_n <= rtp_n) {

			/**
			 * Check fec / (rtp + fec) overhead is less or equal @loss_per_255 / 255.
			 *
			if (255 * fec_n > (rtp_n + fec_n) * loss_per_255) {
				break;
			}*/
			
			/**
			 * Check allowed overhead.
			 */
			overhead = 0;
			if (max_overhead_per_255 && (255 * fec_n > (rtp_n + fec_n) * max_overhead_per_255)) {
				//break;
				overhead = 1;
			}

			/* Probability of 0 loss, i.e. all media packets arrive. */
			p_recovery = ulp_binomial_table[ulp_binomial_table_n_k_2_idx(rtp_n, 0)] * pow(1 - p, rtp_n);
			
			/* Accumulate probability of loss of up to @expected_loss media packets, given all FECs arrive,
			 * i.e. all recoverable scenarios, ignoring some few special cases when recovery is still
			 * possible even if some FECs are missing: so this probability is slightly underestimated
			 * (but simple). */
			i = 1;
			while (i <= ulp_min(fec_n, expected_loss)) {
				p_recovery += ulp_binomial_table[ulp_binomial_table_n_k_2_idx(rtp_n, i)] * pow(p, i) * pow(1 - p, rtp_n - i);// * pow(1 - p, ulp_min(fec_n, expected_loss));
				++i;
			}
			
			//fprintf(stdout, "++ loss=%u: exp_loss=%u, fec_n=%u, rtp_n=%u, p_recovery=%f\toverhead=%u\n", loss_per_255, expected_loss, fec_n, rtp_n, p_recovery, overhead);
			
			if (overhead) {
				break;
			}

			/**
			 * Update optimum if current scheme has better chance of recovery of up to @expected_loss packets
			 * or has same probability of successful recovery but less overhead.
			 */
			if ((p_recovery > p_opt) || (0.0001 > fabs(p_recovery - p_opt) && ((f_n_opt + r_n_opt) * fec_n < (rtp_n + fec_n) * f_n_opt))) {
			
				p_opt = p_recovery;
				f_n_opt = fec_n;
				r_n_opt = rtp_n;
			}

			++fec_n;
		}

		++rtp_n;
	}

ulp_opt_fec_scheme_done:

	*fec_n_opt = f_n_opt;
	*rtp_n_opt = r_n_opt;

	return p_opt;
}


/* For testing. */

/* @payload_len must include CSRC and extensions if any. */
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_gen_rtp(uint8_t v, uint8_t p, uint8_t x, uint8_t cc, uint8_t m, uint8_t pt, uint16_t seq, uint32_t ts, uint32_t ssrc, uint16_t payload_len)
{
	ulp_rtp_pkt_t *pkt = malloc(sizeof(ulp_rtp_hdr_t) + payload_len);
	if (!pkt) {
		return NULL;
	}

	/* memset with the payload_len, better than memsetting with 0. this is for the unit tests (payload) */
	memset(pkt, payload_len, sizeof(ulp_rtp_hdr_t) + payload_len);

	pkt->hdr.v = v;
	pkt->hdr.p = p;
	pkt->hdr.x = x;
	pkt->hdr.cc = cc;
	pkt->hdr.m = m;
	pkt->hdr.pt = pt;
	pkt->hdr.seq = htons(seq);
	pkt->hdr.ts = htonl(ts);
	pkt->hdr.ssrc = htonl(ssrc);

	return pkt;
}

SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_gen_rtp_host(uint8_t v, uint8_t p, uint8_t x, uint8_t cc, uint8_t m, uint8_t pt, uint16_t seq, uint32_t ts, uint32_t ssrc, uint16_t payload_len)
{
	ulp_rtp_pkt_t *pkt = malloc(sizeof(ulp_rtp_hdr_t) + payload_len);
	if (!pkt) {
		return NULL;
	}

	memset(pkt, 0, sizeof(ulp_rtp_hdr_t) + payload_len);

	pkt->hdr.v = v;
	pkt->hdr.p = p;
	pkt->hdr.x = x;
	pkt->hdr.cc = cc;
	pkt->hdr.m = m;
	pkt->hdr.pt = pt;
	pkt->hdr.seq = seq;
	pkt->hdr.ts = ts;
	pkt->hdr.ssrc = ssrc;

	return pkt;
}

SWITCH_DECLARE(int) ulp_test_bytes(char *x, char *y, int n)
{
	int i = 0;

	while (i < n && (x[i] == y[i])) {

		++i;
	}

	return i;
}

SWITCH_DECLARE(void) ulp_die(const char *reason, const char *file, int line)
{
	fprintf(stderr, "Test failed: %s. %s:%d\n", reason, file, line);
	exit(EXIT_FAILURE);
}
